/* $Id: acre.h,v 1.1 2004/04/30 05:16:54 Joel M. Malard Exp d3j540 $ */

/* Include file */
#include "petsc.h"
#include "petscmat.h"
#include "petscksp.h"
#include <math.h>

/* defines */
#define ACRE_SET_CONTEXT		0
#define ACRE_SET_CHOL		1
#define ACRE_GRAD_BACK		2
#define ACRE_MATRIX_COMBINE	2
#define ACRE_VECTOR_COMBINE	3
#define ACRE_VECTOR_BROADCAST	6
#define ACRE_LIKE		4
#define ACRE_GRAD		5
#define ACRE_HESS		7
#define ACRE_LK			8
#define ACRE_SCATTER		9

#define ACRE_GENERAL		9
#define ACRE_TRIANGULAR		10
#define ACRE_DEFAULT		11
#define ACRE_BS			12
#define ACRE_SPAI		13

#define ACRE_FIRST		0
#define ACRE_NEXT		1
#define ACRE_SLIMIT		10

#define ACRE_ML			0
#define ACRE_REML		1

#define ACRE_CRAMER_GRADIENT	0
#define ACRE_ADJOINT_GRADIENT	1

#define ACRE_TRANSPOSE		0
#define ACRE_NO_TRANSPOSE	1
#define ACRE_EPSILON		1.0e-20
#define ACRE_SMALL		1.0e-6
#define ACRE_MAT_HUGE		1000

/* #define HELLO(NAME) { \
      int ierr, rank; \
      ierr = MPI_Comm_rank(PETSC_COMM_WORLD,&rank); CHKERRQ(ierr); \
      PetscSynchronizedPrintf(PETSC_COMM_SELF,"{%d}\t%s\n",rank,NAME); \
      PetscSynchronizedFlush(PETSC_COMM_SELF); \
  }
*/

#define HELLO(NAME)

/* User-defined application contexts */

typedef struct {
	/* problem sizes */
	int	m;	/* Number of equations */
	int	n;	/* Number of parameters 
	   		in the linear model */
	int	nq;	/* Number of variance components */

	/* problem definition */
	Mat	A;	/* Data matrix */
	Vec	b;	/* Response vector */
	Mat	sL;	/* Sparsity structure of the
	   		matrix L */
	int	bTriangular;
	Mat	tD; 	/* The structure of the design matrix */
	Vec	sLDiag;	/* number of diagonal occurance */

	int	ofcn;	/* whether ML or REML */

	int	qc;	/* the process grid has qc columns along which
			 * the columns of A, B and Q are block distributed.
			 * Row-wise distribution of these matrices
			 * are according to Petsc default distribution.
	 		 */
	MPI_Comm	rcomm;	/* communicator of processes assigned to
	 			* the same rows of L, Q, A, etc.
	 			*/
	int	*rstart;	/* first column of A on each process. */
	int	rn;		/* number of columns on this process */

	/* PETSc contexts */
	int	singf;		/* Set to 1 in acre_set_cholesky when L 
	   			is singular, the following SLES
	   			contexts are then meaningless */
	KSP	sles;		/* Linear system solver context */
#if !defined(USE_PETSC_TRANSPOSE)
	KSP	stes;		/* Transpose linear system solver
	   			* context
				*/
#endif

	/* preallocated matrices and vectors */
	Mat	L;		/* Cholesky factor of the 
	   			 * covariance matrix
				 */ 
	Mat	sLTr;		/* To avoid computing a transpose */
	Mat	LTr;		/* Transpose of the Cholesky factor
	   			of the covariance matrix */ 
	Mat	LDot, LDotTr;
	int	k;
	Mat	B;		/* inv(L)*A */
	Mat	Q;		/* Column range of inv(L)*A */
	Mat	R;		/* Cholesky Factor of inv(L)*A */
	Vec	w;		/* w = inv(L)b */
	Vec	v;		/* L^Tu of Ax+Cu=b */
	Vec	u;		/* u of Ax+Cu=b */
	Vec	pVec;		/* Last vector used to assemble L */
	VecScatter	*vScts;	/* VecScatter objects for multiplying
	   			 * by the derivative of L
				 */
	int	bgSet;
	Vec	gVec;		/* gradient vector for pVec if any */

	int	bfSet;
	PetscScalar	fScal;	/* function value for pVec if any */
	PetscScalar	tol;	/* tolerance (relative & absolute for 
				 * TAO solver.
				 */
	IS	*fiS, *tIS;

	char	*plfn;

	/* Profiling */
	int	events[20];

} acre_cntxt;

/* BLAS declarations modified from petscblaslapack.h,v 1.39 2000/05/10
 * 16:44:31 bsmith Exp $ */

#if !defined(DGER)

/* This include file on the Cray T3D/T3E defines the interface between 
 * Fortran and C representations of character strings.
 */
#if defined(PETSC_USES_CPTOFCD)
#include <fortran.h>
#endif

#if !defined(PETSC_USE_COMPLEX)

/* These are real case with no character string arguments */

#if defined(PETSC_USES_FORTRAN_SINGLE) 
/* For these machines we must call the single precision Fortran version */
#define DGER   SGER
#endif

#if defined(PETSC_USE_SINGLE)

#if defined(PETSC_HAVE_FORTRAN_UNDERSCORE) || defined(PETSC_BLASLAPACK_F2C)
#define BLger_ sger_
#elif defined(PETSC_HAVE_FORTRAN_CAPS)
#define BLger_ SGER
#else
#define BLger_ sger
#endif

#else

#if defined(PETSC_HAVE_FORTRAN_UNDERSCORE) || defined(PETSC_BLASLAPACK_F2C)
#define BLger_ dger_
#elif defined(PETSC_HAVE_FORTRAN_CAPS)
#define BLger_ DGER
#else
#define BLger_ dger
#endif

#endif

/* Real with character string arguments.  */
#if defined(PETSC_USES_CPTOFCD)
/* Note that this assumes that machines which use cptofcd() use 
 * the PETSC_HAVE_FORTRAN_CAPS option. This is true on the Cray T3D/T3E.
 */
#define BLger_ DGER

#elif defined(PETSC_USE_SINGLE)

#if defined(PETSC_HAVE_FORTRAN_UNDERSCORE) || defined(PETSC_BLASLAPACK_F2C)
#define BLger_ sger_
#elif defined(PETSC_HAVE_FORTRAN_CAPS)
#define BLger_ SGER
#else
#define BLger_ sger
#endif

#else /* PETSC_USE_SINGLE */

#if defined(PETSC_HAVE_FORTRAN_UNDERSCORE) || defined(PETSC_BLASLAPACK_F2C)
#define BLger_ dger_
#elif defined(PETSC_HAVE_FORTRAN_CAPS)
#define BLger_ DGER
#else
#define BLger_  dger
#endif

#endif /* PETSC_USES_CPTOFCD */

#else
/* Complex with no character string arguments */
#if defined(PETSC_USES_FORTRAN_SINGLE)
#define ZGER  CGER
#endif

#if defined(PETSC_HAVE_FORTRAN_UNDERSCORE) || defined(PETSC_BLASLAPACK_F2C)
#define BLger_ zger_
#elif defined(PETSC_HAVE_FORTRAN_CAPS)
#define BLger_ ZGER
#else
#define BLger_ zger
#endif
    
#if defined(PETSC_USES_CPTOFCD)
#define BLger_ ZGER
#else
#define BLger_  zger
#endif

#endif

EXTERN_C_BEGIN

/* 
   	BLdot cannot be used with COMPLEX because it cannot 
   	handle returing a double complex to C++.
*/
EXTERN void BLger_(int*,int*,PetscScalar*,PetscScalar*,int*,PetscScalar*,int*,PetscScalar*,int*);

EXTERN_C_END

#endif

/* Prototypes */
#include "tao.h"
int acre_adjoint_function_gradient( TAO_APPLICATION reml, Vec iVec, double *fVal, Vec gVec, void *cntxt);
int acre_adjoint_gradient( TAO_APPLICATION reml, Vec iVec, Vec gVec, void * cptr);
int acre_bcast_vector(Vec iVec, Vec iSeq);
int acre_close_scatter( acre_cntxt *cntxt, int p, Vec u, Vec v, IS *fiS, IS *tiS, VecScatter *iScatter);
int acre_cramer_function_gradient( TAO_APPLICATION reml, Vec iVec, double *fVal, Vec gVec, void *cntxt);
int acre_cramer_gradient( TAO_APPLICATION reml, Vec iVec, Vec gVec, void * cptr);
int acre_detl(Mat sL, Vec sLDiag);
int acre_effects( acre_cntxt *cntxt, Vec x, PetscScalar *rnorm);
int acre_end(acre_cntxt *cntxt);
int acre_fd_backward(acre_cntxt *cntxt, Mat F);
int acre_fdl_times_vec( acre_cntxt *cntxt, int p, Vec u, Vec v, int flag, IS *fiS, IS *tiS, VecScatter *iScatter);
int acre_get_dense_matrix_column_vector(Mat A, Vec yy, int col);
int acre_likelihood ( TAO_APPLICATION reml, Vec iVec, PetscScalar *fVal, void * cptr);
int acre_matrix_combine(acre_cntxt *cntxt, Mat AL, Mat AG );
int acre_matrix_view(acre_cntxt *cntxt, Mat A , char *msg);
int acre_mgs(acre_cntxt *cntxt, Mat Q, Mat R);
int acre_set_cholesky(acre_cntxt *cntxt, Vec iVec);
int acre_set_context(acre_cntxt *cntxt, Vec iVec);
int acre_setup_scatter( acre_cntxt *cntxt, int p, Vec u, Vec v, IS *fiS, IS *tiS, VecScatter *iScatter);
int acre_sles_mat( acre_cntxt *cntxt, Mat A, Mat B, Vec v, int bTranspose);
int acre_start(Mat A, Vec b, Mat sL, Vec c, int qc, acre_cntxt *cntxt);
int acre_vector_broadcast(acre_cntxt *cntxt, Vec u, int root );
int acre_vector_combine(acre_cntxt *cntxt, Vec u );
int acre_write_tao_solver_state(TAO_SOLVER reml);

int acre_print_aij_matrix_serial(Mat A, char * infile, PetscTruth format);
int acre_print_dense_matrix_serial(Mat A, char * infile, PetscTruth format);
int acre_print_vector_serial(Vec b, char *infile, PetscTruth format);
int acre_read_aij_matrix(MPI_Comm comm, char * infile, PetscTruth format, Mat * A);
int acre_read_dense_matrix(MPI_Comm comm, char * infile, PetscTruth format, Mat * A);
int acre_read_vector(MPI_Comm comm, char * infile, PetscTruth format, Vec * x);
int acre_read_vector_serial(MPI_Comm comm, char * infile, PetscTruth format, Vec * x);
