/*$Id: acre_cramer_gradient.c,v 1.1 2004/04/30 J. Malard Exp d3j540 $ */

#include <assert.h>
#include "tao_solver.h"
#include "petscblaslapack.h"
#include "acre.h"
#undef __FUNC__
#include "copyright.h"
#define __FUNC__ "acre_cramer_gradient"
int acre_cramer_gradient( TAO_APPLICATION reml, Vec iVec, Vec gVec,
	void * cptr)
{
	acre_cntxt *cntxt = (acre_cntxt *) cptr;
	int n = cntxt->n;
	int nq  = cntxt->nq;
	int qc = cntxt->qc;
	int rn = cntxt->rn;

	int rs;
	int i, ierr, j, k;
	int here, npes, rank, cank;
	int *rows, nrows, start, end;
	int *rstrt = cntxt->rstart;
	Vec aVec;
	Vec bVec;
	Mat F;
	Mat K;
	PetscScalar *ia, *da;
	PetscScalar zero=0.0;
	PetscScalar wg, wc, *vals;
	PetscScalar * fVals;
	PetscScalar * rVals;
	PetscScalar * uVals;
	PetscScalar * wVals;
	PetscScalar * xVals;
	PetscScalar * kVals;
	PetscScalar * gVals;
	IS fiS;
	IS tiS;
	VecScatter iScatter;
	int inew;
	static int ncall = 0;

	HELLO("acre_cramer_gradient enter");
	ierr = MPI_Comm_rank(cntxt->rcomm,&rank);
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_size(cntxt->rcomm,&npes);
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_rank(PETSC_COMM_WORLD,&cank);
		assert(ierr==MPI_SUCCESS);

	ierr = MPI_Comm_rank(MPI_COMM_WORLD,&here);
		assert(ierr==MPI_SUCCESS);
	
	/* Set up the problem */
	rs = rstrt[rank];

	/* call to acre_set_context */
	inew = acre_set_context(cntxt, iVec);

	if ( inew>0 && cntxt->bgSet!=0 )
	{
		ierr = VecCopy(cntxt->gVec,gVec);
		return 0;
	}
	ierr = VecDuplicate ( cntxt->b, &bVec);			CHKERRQ(ierr);
	
	ierr = VecGetArray( iVec, &ia );			CHKERRQ(ierr);
	ierr = VecGetArray( cntxt->sLDiag, &da );		CHKERRQ(ierr);
	ierr = VecSet( &zero, gVec);				CHKERRQ(ierr);

	/* Det C contribution */
	for ( i=0; i<nq; i++ )
	{
		if ( da[i]!=0 )
		{
			wc = 2*da[i]/ia[i];
			VecSetValue(gVec,i,wc,ADD_VALUES);
		}
	}
	
	ierr = VecRestoreArray( cntxt->sLDiag, &da );		CHKERRQ(ierr);
	ierr = VecRestoreArray( iVec, &ia );			CHKERRQ(ierr);

	/* DetC */

	if ( cntxt->ofcn==ACRE_REML )
	{

		/* Det R contribution */
		ierr = VecDuplicate ( cntxt->b, &aVec);		CHKERRQ(ierr);
		ierr = VecGetOwnershipRange (aVec,
			&start, &end);				CHKERRQ(ierr);
		nrows = end-start;
		rows = NULL;
		if ( nrows!=0 )
		{
			ierr = PetscMalloc(nrows*sizeof(int),
				&rows);				CHKERRQ(ierr);
			for ( i=0; i<nrows; i++ )
				rows[i] = i+start;
		}

       		ierr = MatCreateMPIDense(
			PETSC_COMM_WORLD, PETSC_DECIDE,
			PETSC_DECIDE, cntxt->m, cntxt->rn,
			TAO_NULL, &K);				CHKERRQ(ierr);

		/* allocate storage for F=L^(-T)B */
       		ierr = MatCreateMPIDense(
			PETSC_COMM_WORLD, PETSC_DECIDE,
			PETSC_DECIDE, cntxt->m, cntxt->n,
			TAO_NULL, &F);				CHKERRQ(ierr);

		/* allocate storage for a block row distributed
		 * mxn matrix
		 */	
       		ierr = PetscMalloc(nrows*n*sizeof(PetscScalar),
			&gVals);				CHKERRQ(ierr);
	
		/* detR: allocate */

		/* compute F in L^T*F=B */

		if ( qc>1 )
		{
			ierr = acre_sles_mat(cntxt,cntxt->B,K,
				cntxt->b, ACRE_TRANSPOSE);	CHKERRQ(ierr);
			ierr = acre_matrix_combine( cntxt, K,
				F);				CHKERRQ(ierr);
		}
		else
		{
			ierr = acre_sles_mat(cntxt,cntxt->B,F,
				cntxt->b, ACRE_TRANSPOSE);	CHKERRQ(ierr);
		}
		/* detR: compute F */
	
		/* Compute U in R^T*R*U=I */
		ierr = MatGetArray( cntxt->R, &rVals );		CHKERRQ(ierr);
		ierr = PetscMalloc(n*n*sizeof(PetscScalar),
			&uVals);				CHKERRQ(ierr);
		for ( i=0; i<n; i++ )
		{
			for ( j=0; j<n; j++ )
			{
				uVals[i*n+j] = (i==j) ?
					1.0 : 0.0;
			}
		}
		LAtrtrs_("U","T","N",&n,&n,
			rVals,&n,uVals,&n,&ierr);		CHKERRQ(ierr);
		LAtrtrs_("U","N","N",&n,&n,
			rVals,&n,uVals,&n,&ierr);		CHKERRQ(ierr);
		/* detR: compute U */

		ierr = MatGetArray(F, &fVals);			CHKERRQ(ierr);
		ierr = PetscMalloc(rn*sizeof(double),
			&wVals);				CHKERRQ(ierr);
		ierr = PetscMalloc(rn*sizeof(double),
			&xVals);				CHKERRQ(ierr);

		for ( k=0; k<nq; k++ )
		{

			/* first compute local columns of
			 * L^{k}*B
			 */
			acre_setup_scatter ( cntxt, k,
				aVec, bVec, &fiS,
				&tiS, &iScatter);

			for ( j=0; j<rn; j++ )
			{

				ierr = acre_get_dense_matrix_column_vector(
					cntxt->B, aVec, j);	CHKERRQ(ierr);

				/* Compute L^{k}*B(:,j) */
				acre_fdl_times_vec ( cntxt,
					k, aVec, bVec,
					ACRE_NO_TRANSPOSE,
					&fiS, &tiS,
					&iScatter);

				/* save in K */
				if ( nrows!=0 )
				{
					ierr = VecGetArray(bVec,
						&vals);		CHKERRQ(ierr);
					ierr = MatSetValues( K,
						nrows, rows, 1,
						&j, vals,
						INSERT_VALUES);	CHKERRQ(ierr);
					ierr = VecRestoreArray(
						bVec, &vals);	CHKERRQ(ierr);
				}
			}

			ierr = MatAssemblyBegin(K,
				MAT_FINAL_ASSEMBLY);		CHKERRQ(ierr);
			ierr = MatAssemblyEnd(K,
				MAT_FINAL_ASSEMBLY);		CHKERRQ(ierr);

			acre_close_scatter ( cntxt, k,
				aVec, bVec,
				&fiS, &tiS,
				&iScatter);
			/* detR: main -> compute K */

			/* compute local contribution to
			 * L^{k}B*U
			 */
			ierr = MatGetArray(K,&kVals);		CHKERRQ(ierr);
			{
			PetscScalar aone = 1.0;
			PetscScalar azero = 0.0;
			BLgemm_("NoTr", "NoTr", &nrows, &n, &rn,
				&aone, kVals, &nrows, &uVals[rs],
				&n, &azero, gVals, &nrows);
			}
			ierr = MatRestoreArray(K,&kVals);	CHKERRQ(ierr);

			/* detR: main -> K*U */

			/* compute local contribution to
			 * vec(F)^T*vec(L^{k}B*U) - note 
			 * the use of U's symmetry.
			 */
			{
			int ione = 1;
			int ne = nrows*n;
			wc = BLdot_( &ne, fVals, &ione,
				gVals, &ione);
			}
			/* detR: main -> local vec{F}^T*vec{K*U} */

			/* sum up local contributions across
			 * columns of the process grid.
			 */
			if ( qc>1 )
			{
				ierr = MPI_Allreduce(
					&wc, &wg, 1,
					MPI_DOUBLE,
					MPI_SUM,
					cntxt->rcomm);	assert(ierr==MPI_SUCCESS);
				wc = wg;
			}
			ierr = MPI_Allreduce(
				&wc, &wg, 1,
				MPI_DOUBLE, MPI_SUM,
				PETSC_COMM_WORLD);	assert(ierr==MPI_SUCCESS);
			wg *= -2;
			VecSetValue( gVec, k, wg, ADD_VALUES);

			/* detR: main -> global sum */
		}
		ierr = MatRestoreArray(F, &fVals);			CHKERRQ(ierr);

		PetscFree(gVals);
		PetscFree( wVals );
		PetscFree( xVals );
		PetscFree ( uVals );
		PetscFree ( rows );

		ierr = VecDestroy(aVec);		CHKERRQ(ierr);
		ierr = MatDestroy(F);			CHKERRQ(ierr);
		ierr = MatDestroy(K);			CHKERRQ(ierr);
		/* detR: main loop */
	}

	/* Residual Term */
	ierr = PetscMalloc(nq*sizeof(double),&vals);	CHKERRQ(ierr);
	for ( i=0; i<nq; i++ ) vals[i] = 0;
	
	for ( i=0; i<nq; i++ )
	{

		if ( i%qc!=rank) continue; 

		acre_setup_scatter ( cntxt, i, cntxt->u,
			bVec, &fiS, &tiS,
			&iScatter);
		acre_fdl_times_vec( cntxt, i, cntxt->u,
			bVec, ACRE_TRANSPOSE,&fiS,
			&tiS, &iScatter);
		acre_close_scatter ( cntxt, i, cntxt->u,
			bVec, &fiS, &tiS,
			&iScatter);

		ierr = VecDot( cntxt->v, bVec, &wc );	CHKERRQ(ierr);
		vals[i] = -2*wc;
	}
	/* res: scatter */

	ierr = PetscMalloc(nq*sizeof(int),&rows);	CHKERRQ(ierr);
	for ( i=0; i<nq; i++ ) rows[i] = i;

	/* add together local contributions */
	if ( qc>1 )
	{
		double *wals;
	       	ierr = PetscMalloc(nq*sizeof(double),
			&wals);				CHKERRQ(ierr);

		ierr = MPI_Allreduce( vals, wals,
			nq, MPI_DOUBLE, MPI_SUM,
			cntxt->rcomm);
			assert(ierr==MPI_SUCCESS);
		ierr = VecSetValues( gVec, nq, rows,
			wals, ADD_VALUES);		CHKERRQ(ierr);

		PetscFree(wals);
	}
	else
	{
		ierr = VecSetValues( gVec, nq, rows,
			vals, ADD_VALUES);		CHKERRQ(ierr);
	}
	PetscFree ( vals );
	PetscFree ( rows );
	/* res: reduce */
	
	/* Free remaining matrices and vectors */
	ierr = VecDestroy(bVec);			CHKERRQ(ierr);
	
	/* Sum all contributions */
	
	ierr = VecAssemblyBegin(gVec);			CHKERRQ(ierr);
	ierr = VecAssemblyEnd(gVec);			CHKERRQ(ierr);

	ierr = VecCopy(gVec,cntxt->gVec);		CHKERRQ(ierr);
	cntxt->bgSet = 1;
	
	ierr = VecGetArray(gVec,&vals);			CHKERRQ(ierr);
	ierr = MPI_Bcast(vals, nq, MPI_DOUBLE, 0,
		MPI_COMM_WORLD);
		assert(ierr==MPI_SUCCESS);
	ierr = VecRestoreArray(gVec,&vals);		CHKERRQ(ierr);
	
	ncall++;

	HELLO("acre_cramer_gradient exit");
	return 0;
}
