/*$Id: acre_get_dense_matrix_column_vector.c 1.1 2004/04/30 J. Malard d3j540 $*/

#include "tao_solver.h"
#include "petscblaslapack.h"
#include "petsclog.h"
#include "acre.h"
#include "petsclog.h"
#include <assert.h>
#include <mpi.h>
#include <math.h>
#include "copyright.h"
/*
	acre_get_dense_matrix_column_vector - Gets the values from a
	given column of a dense MPI matrix.

	Collective on Mat and Vec

	Input Parameters:
		A - the matrix
		yy - the vector
		col - the column requested

	Modified from MatGetColumnVector.

*/
#undef __FUNC__  
#define __FUNC__ "acre_get_dense_matrix_column_vector"
int acre_get_dense_matrix_column_vector(Mat A, Vec yy, int col)
{
	PetscScalar *y,zero = 0.0;
	int ierr,i,M,N,Mv,Rs,Re,rs,re;
	int nrows;
	PetscScalar *vals;

	HELLO("acre_get_dense_matrix_column_vector enter");  

	ierr = MatGetSize(A,&M,&N);			CHKERRQ(ierr);
	if (col < 0) 
	{
		SETERRQ1( PETSC_ERR_ARG_WRONG,
		"Requested negative column: %d",col);
	}
	if (col >= N)
	{
		SETERRQ2( PETSC_ERR_ARG_WRONG,
		"Requested column %d larger than number columns in \n\
		matrix %d",col,N);
	}

	ierr = VecGetSize(yy,&Mv);			CHKERRQ(ierr);
	if (M != Mv)
	{
		SETERRQ2( PETSC_ERR_ARG_WRONG,
		"Matrix does not have same number of \n\
		columns %d as vector %d",M,Mv);
	}

	ierr = MatGetOwnershipRange(A,&Rs,&Re);		CHKERRQ(ierr);
	ierr = VecGetOwnershipRange(yy,&rs,&re);	CHKERRQ(ierr);
	if (Rs != rs || Re != re)
	{
		SETERRQ4(
		PETSC_ERR_ARG_WRONG,"Matrix %d %d \n\
		does not have same ownership range as vector %d %d",
		Rs,Re,rs,re);
	}

	ierr = VecSet(&zero,yy);			CHKERRQ(ierr);
	ierr = VecGetArray(yy,&y);			CHKERRQ(ierr);

	nrows = Re-Rs;

	ierr = MatGetArray(A,&vals);			CHKERRQ(ierr);

	for ( i=0; i<nrows; i++ )
	{
		y[i] = vals[col*nrows+i];
	}

	ierr = MatRestoreArray(A,&vals);		CHKERRQ(ierr);
	ierr = VecRestoreArray(yy,&y);			CHKERRQ(ierr);  

	HELLO("acre_get_dense_matrix_column_vector exit");  

	return 0;

}
