/*$Id: acre_read_dense_matrix.c,v 1.1 2004/04/30 Joel M. Malard Exp d3j540 $ */

#include <stdlib.h>
#include <assert.h>
#include "tao.h"
#include "acre.h"
#include "petscmat.h" 
#include "copyright.h"
#undef __FUNC__
#define __FUNC__ "acre_read_dense_matrix"

int acre_read_dense_matrix(MPI_Comm comm, char * infile, PetscTruth format,
	Mat * A)
{
	int ierr;

	HELLO("acre_read_dense_matrix enter");

	if ( format==PETSC_TRUE )
	{
		int i, j;
		int m, n;
		FILE *fp;
		int *rows;
		PetscScalar *vals;
		fp = fopen(infile,"r");
			assert(fp!=NULL);
		ierr = fscanf(fp, "%d\t%d\n",&m,&n);
			assert(ierr==2);

		ierr = MatCreateMPIDense( comm,
			PETSC_DECIDE, PETSC_DECIDE,
			m, n, PETSC_NULL, A);			CHKERRQ(ierr);

		ierr = PetscMalloc(m*sizeof(PetscScalar),
			&vals);					CHKERRQ(ierr);
		ierr = PetscMalloc(m*sizeof(int),&rows);	CHKERRQ(ierr);

		for ( i=0; i<n; i++ )
		{
			for ( j=0; j<m; j++ )
			{
				rows[j] = j;
				ierr = fscanf(fp,"%lg",
					&vals[j]);
					assert(ierr==1);
			}
			ierr = MatSetValues(*A, m, rows,
				1, &i, vals, INSERT_VALUES);	CHKERRQ(ierr);
		};
		ierr = MatAssemblyBegin(*A,MAT_FINAL_ASSEMBLY);	CHKERRQ(ierr);
		ierr = MatAssemblyEnd(*A,MAT_FINAL_ASSEMBLY); 	CHKERRQ(ierr);

		ierr = PetscFree(vals);				CHKERRQ(ierr);
		ierr = PetscFree(rows);				CHKERRQ(ierr);
		fclose(fp);
	}
	else
	{
		PetscViewer  fd;
		ierr = PetscViewerBinaryOpen(PETSC_COMM_WORLD,
			infile, PETSC_FILE_RDONLY,&fd);		CHKERRQ(ierr);
		ierr = MatLoad(fd, MATMPIDENSE, A);		CHKERRQ(ierr);
		ierr = PetscViewerDestroy(fd);			CHKERRQ(ierr);
	}

	HELLO("acre_read_dense_matrix exit");

	return ierr;
}
