/*$Id: acre_set_cholesky.c,v 1.1 2004/04/30 Joel M. Malard Exp d3j540 $ */

#include "tao_solver.h"
#include "petscblaslapack.h"
#include "petsclog.h"
#include "acre.h"
#include "petsclog.h"
#include <assert.h>
#include <mpi.h>
#include "copyright.h"
#undef __FUNC__
#define __FUNC__ "acre_set_cholesky"
int acre_set_cholesky(acre_cntxt *cntxt, Vec iVec)
{
	/* Purpose:
	 * Initialize the distributed matrix cntxt->L and cntxt->LTr
	 * using the parameters in the sequential vector iVec.
	 */
	int ierr, start, end, i, j;
	int ncols, *theta, *icols, nq;
	PetscScalar *vals, *vcols, *vc;

	HELLO("acre_set_cholesky enter");

	/* Get a local copy of the input vector */
	ierr = VecGetSize(iVec, &nq);				CHKERRQ(ierr);
	ierr = VecGetArray(iVec,&vc);				CHKERRQ(ierr);

	/* Fill in the Cholesky factor of the covariance matrix */
	ierr = PetscMalloc(2*nq*sizeof(PetscScalar),&vals);	CHKERRQ(ierr);
	ierr = PetscMalloc(2*nq*sizeof(int),&icols);		CHKERRQ(ierr);
	ierr = MatGetOwnershipRange(cntxt->sL,&start,&end);	CHKERRQ(ierr);
	for ( i=start; i<end; i++ )
	{
		ierr = MatGetRow(cntxt->sL, i, &ncols, &theta,
			&vcols);				CHKERRQ(ierr);
		if ( ncols!=0 )
		{
			for ( j=0; j<ncols; j++ )
			{ 
				vals[j] = vc[theta[j]];
				icols[j] = ((int)vcols[j])-1;
			}
			ierr = MatSetValues(cntxt->L, 1, &i,
				ncols, icols, vals,
				INSERT_VALUES);			CHKERRQ(ierr);
			ierr = MatSetValues(cntxt->LTr, ncols,
				icols, 1, &i, vals,
				INSERT_VALUES);			CHKERRQ(ierr);
		}
		ierr = MatRestoreRow(cntxt->sL, i, &ncols,
				&theta, &vcols);		CHKERRQ(ierr);
	}

	ierr = VecRestoreArray(iVec,&vc);			CHKERRQ(ierr);

	ierr = MatAssemblyBegin(cntxt->L,MAT_FINAL_ASSEMBLY);	CHKERRQ(ierr);
	ierr = MatAssemblyEnd(cntxt->L,MAT_FINAL_ASSEMBLY);	CHKERRQ(ierr);
	ierr = MatAssemblyBegin(cntxt->LTr,
		MAT_FINAL_ASSEMBLY);				CHKERRQ(ierr);
	ierr = MatAssemblyEnd(cntxt->LTr,MAT_FINAL_ASSEMBLY);	CHKERRQ(ierr);

	PetscFree(vals);
	PetscFree(icols);

	HELLO("acre_set_cholesky exit");

	return 0;

}
